library dom;
import 'dart:collection';
import 'dart:convert';
import 'dart:html';
import 'package:domquery/dom/exception.dart' as exception;
import 'package:domquery/dom/utils.dart';

/**
 * One Class to rule them all, One Class to find them, 
 * One Class to bring them all and in the darkness bind them.
 * 
 * This Class implements the "God object" antipattern:
 * http://en.wikipedia.org/wiki/God_object
 * 
 * The Class is intented for all those programmers who "don't care".
 */
class DomNode extends IterableBase {
  /**
   * This constant is used by the data() method.
   */
  static const _DATA_NAMESPACE = '__dartQuery__';
  
  /**
   * List of elements.
   */
  List<Element> _elements = [];
  
  /**
   * Creates a node.
   */
  DomNode(String nodeName, {Document context, Map attributes, Object text, Object html, void callback(DomNode target)}) {
    if (context == null) {
      context = document;
    }
    
    var elem = context.createElement(nodeName);
    _elements.add(elem);
    
    if (attributes != null) {
      attributes.forEach((String name, Object value) {
        attr(name, value);
      });
    }
    
    if (text != null) {
      this.text(text);
    }
    
    if (html != null) {
      this.html(html);
    }
    
    if (callback != null) {
      Function.apply(callback, [this]);
    }
  }
  
  /**
   * Creates a node from a string.
   */
  DomNode.fromString(String str, [String contentType = 'text/xml']) {
    var parser = new DomParser();
    var doc = parser.parseFromString(str, contentType);
    
    _elements = [doc.documentElement];
  }
  
  /**
   * Creates a node from a document.
   */
  DomNode.fromDocument(Document doc) {
    _elements = [doc.documentElement];
  }
  
  /**
   * Creates a node from an element.
   */
  DomNode.fromElement(Element element) {
    _elements = [element];
  }
  
  /**
   * Creates a node from a list.
   */
  DomNode.fromList(List<Element> elements) {
    _elements = elements;
  }
  
  /**
   * Implements IterableBase.iterator
   */
  Iterator<DomNode> get iterator {
    var ret = [];
    var len = _elements.length;
    
    for (var i = 0; i < len; i++) {
      ret.add(new DomNode.fromElement(_elements[i]));
    }
    
    return ret.iterator;
  }
  
  /**
   * Searchs elements by a CSS selector.
   */
  DomNode query(String selectors) {
    var elements = [];
    
    _elements.forEach((element) {
      var items = element.querySelectorAll(selectors);
      
      items.forEach((Element item) {
        if (!elements.contains(item)) {
          elements.add(item);
        }
      });
    });
    
    return new DomNode.fromList(elements);
  }
  
  /**
   * Gets the native element.
   */
  Element element() {
    return _elements.length > 0? _elements[0] : null;
  }
  
  /**
   * Gets the node name.
   */
  String name() {
    return _elements.length > 0? _elements[0].nodeName : '';
  }
  
  /**
   * Gets or sets an attribute.
   */
  String attr(String name, [Object value]) {
    if (value != null) {
      _setAttribute(name, value.toString());
    }
    
    return _getAttribute(name);
  }
  
  /**
   * Gets or sets a CSS attribute.
   */
  String css(String name, [String value]) {
    if (value != null) {
      _setCssAttribute(name, value);
    }
    
    return _getCssAttribute(name);
  }
  
  /**
   * Gets or sets inner html.
   */
  String html([Object value]) {
    if (value != null) {
      _setInnerHtml(value);
    }
    
    return _getInnerHtml();
  }
  
  /**
   * Gets or sets inner text.
   */
  String text([Object value]) {
    if (value != null) {
      _setInnerText(value);
    }
    
    return _getInnerText();
  }
  
  /**
   * Gets or sets data.
   */
  Object data(String name, [Object value]) {
    if (value != null) {
      _setData(name, value);
    }
    
    return _getData(name);
  }
  
  /**
   * Gets the parent element.
   * 
   * This function may return null if the element has no parent.
   */
  DomNode parent() {
    return _elements.length > 0? new DomNode.fromElement(_elements[0].parent): null;
  }
  
  /**
   * Appends content.
   */
  void append(Object obj) {
    _elements.forEach((element) {
      element.insertAdjacentHtml('beforeend', obj.toString());
    });
  }
  
  /**
   * Prepends content.
   */
  void prepend(Object obj) {
    _elements.forEach((element) {
      element.insertAdjacentHtml('afterbegin', obj.toString());
    });
  }
  
  /**
   * Removes all child nodes.
   */
  void clean() {
    _elements.forEach((element) {
      while (element.hasChildNodes()) {
        element.firstChild.remove();
      }
    });
  }
  
  /**
   * Removes elements.
   */
  void remove() {
    while (_elements.length > 0) {
      var element = _elements.removeLast();
      
      element.remove();
    }
  }
  
  /**
   * Gets an attribute.
   */
  String _getAttribute(String name) {
    return _elements.length > 0? _elements[0].getAttribute(name) : '';
  }
  
  /**
   * Sets an attribute.
   */
  void _setAttribute(String name, Object value) {
    _elements.forEach((element) {
      element.setAttribute(name, value.toString());
    });
  }
  
  /**
   * Sets a CSS attribute.
   */
  void _setCssAttribute(String name, String value) {
    _elements.forEach((element) {
      var attrs = _getCssAttributesMap(element);
      
      attrs[name] = value;
      _setCssAttributesMap(element, attrs);
    });
  }
  
  /**
   * Gets a CSS attribute.
   */
  String _getCssAttribute(String name) {
    var ret = '';
    
    if (_elements.length > 0) {
      var attrs = _getCssAttributesMap(_elements[0]);
      
      name = name.toLowerCase();
      if (attrs.containsKey(name)) {
        ret = attrs[name];
      }
    }
    
    return ret;
  }
  
  /**
   * Gets the list of attributes of an element.
   */
  Map<String, String> _getCssAttributesMap(Element element) {
    var ret = {};
    var css = element.getAttribute('style');
    
    if (css != null) {
      css.split(';').forEach((cssAttr) {
        var item = cssAttr.split(':');
        var key = item.length > 0? item[0].trim().toLowerCase() : '';
        var value = item.length > 1? item[1].trim() : '';
        
        if (key.length > 0 && value.length > 0) {
          ret[key] = value;
        }
      });
    }
    
    return ret;
  }
  
  /**
   * Sets the list of attributes of an element.
   */
  void _setCssAttributesMap(Element element, Map<String, String> attrs) {
    var arr = [];
    
    attrs.keys.forEach((key) {
      if (key != null && key.length > 0) {
        arr.add(key.toLowerCase() + ': ' + attrs[key]);
      }
    });
    
    attr('style', arr.join(';'));
  }
  
  /**
   * Gets inner html.
   */
  String _getInnerHtml() {
    var ret = new StringBuffer();
    
    if (_elements.length > 0) {
      var nodes = _elements[0].childNodes;
      
      nodes.forEach((node) {
        ret.write(node2str(node));
      });
    }
    
    return ret.toString();
  }
  
  /**
   * Sets inner html.
   */
  void _setInnerHtml(Object value) {
    clean();
    _elements.forEach((element) {
      element.appendHtml(value.toString());
    });
  }
  
  /**
   * Gets inner text.
   */
  String _getInnerText() {
    return _elements.length > 0? _elements[0].text : '';
  }
  
  /**
   * Sets inner text.
   */
  void _setInnerText(Object value) {
    var escape = const HtmlEscape(HtmlEscapeMode.ELEMENT);
    
    clean();
    _elements.forEach((element) {
      element.text = escape.convert(value.toString());
    });
  }
  
  /**
   * Gets the data object of an element.
   */
  Map _getDataObject(Element element) {
    var attr = element.getAttribute(_DATA_NAMESPACE);
    var data = attr != null? JSON.decode(attr): {};
    
    if (data is! Map) {
      throw new exception.DomException('The node doesn\'t contain valid data');
    }
    
    return data;
  }
  
  /**
   * Sets the data object of an element.
   */
  void _setDataObject(Element element, Map data) {
    element.setAttribute(_DATA_NAMESPACE, JSON.encode(data));
  }
  
  /**
   * Gets data.
   */
  Object _getData(String name) {
    var str = null;
    
    if (_elements.length > 0) {
      var data = _getDataObject(_elements[0]);
      
      if (data.containsKey(name)) {
        str = data[name];
      }
    }
    
    return str;
  }
  
  /**
   * Sets data.
   */
  void _setData(String name, Object value) {
    _elements.forEach((element) {
      var data = _getDataObject(element);
      
      data[name] = value;
      _setDataObject(element, data);
    });
  }
  
  /**
   * Gets a string representation of the object.
   */
  String toString() {
    var str = new StringBuffer();
    
    _elements.forEach((element) {
      str.write(node2str(element));
    });
    
    return str.toString();
  }
}